/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "ap-triggers.hh"
#include "xsil/MetaIO.hh"
#include "framedir.hh"
#include "DVecType.hh"
#include <iostream>
#include <fstream>
#include <sstream>

using namespace std;

//=======================================  Trigger class constructor
triggers::triggers(void)
   : _debug(0), _total(0), _nVetoed(0)
{
}

//=======================================  Trigger class constructor
triggers::~triggers(void) {
}

//=======================================  Read trigger data
void
triggers::add_select(const trig_select& ts) {
   _trig_select_list.push_back(ts);
}

//=======================================  Read trigger data
void
triggers::add_veto(const trig_veto& tv) {
   _trig_veto_list.push_back(tv);
}

//=======================================  Read trigger data
void
triggers::read_triggers(const std::vector<std::string>& file,
			gps_t start, gps_t stop, Interval dt) {
   size_t nRead(0), nTime(0), nSelect(0);

   //-----------------------------------  Throw away any previous data
   gps_t prev_start = _series.getStartTime().getS();
   if (prev_start < start) {
      _series.eraseStart(start - prev_start);
   }
   if (stop <= start) return;

   //-----------------------------------  Zero out new time range.
   gps_t previous = start;
   if (_series.empty()) {
      int zero = 0;
      unique_ptr<DVectI> dv(new DVectI(1, &zero));
      _series.setData(Time(start), dt, dv.release());
   } else {
      previous = _series.getEndTime().getS();
   }
   _series.extend(Time(stop));
   _total = size_t(_series.getAverage() * _series.getNSample() + 0.5);

   //-----------------------------------  Get a list of xml files.
   FrameDir fd;
   try {
      for (int i=0; i<file.size(); i++) fd.add(file[i].c_str(), true);
   } catch (std::exception& e) {
      cerr << "triggers::read_triggers failed: " << e.what() << endl;
      return;
   }

   //-----------------------------------  Make sure trigger files were found
   gps_t beg_time = fd.begin()->getStartGPS();
   if (beg_time >= stop) return;
   
   gps_t end_time = (--fd.end())->getEndGPS();
   if (start >= end_time) return;
   auto endfile = fd.find(Time(stop));
   if (endfile != fd.end()) endfile++;
   auto begfile = fd.begin();
   if (start > beg_time) begfile = fd.find(Time(start));
   if (_debug > 1) cout << "first trigger file: "<< begfile->getFile()
			<< " last file: " << endfile->getFile() << endl;

   DVectI& dvi = dynamic_cast<DVectI&>(*_series.refDVect()); 
   for (FrameDir::file_iterator i=begfile; i != endfile; ++i) {
      if (_debug > 1) cout << "Read trigger file: " << i->getFile() << endl;
      xsil::MetaIO mio(i->getFile().c_str(), _table_name.c_str());
      if (!mio.is_open()) {
	 cerr << "Unable to open trigger file: " << i->getFile()
	      << " table: " << _table_name << endl;
	 continue;
      }
      
      while (mio.getRow()) {
	 nRead++;
	 Time::ulong_t gps_s  = mio.getInt(_time_field.c_str());
	 if (gps_s < previous || gps_s >= stop) continue;
	 nTime++;
	 Time::ulong_t gps_ns = mio.getInt(_time_ns_field.c_str());
	 if (select_trigger(mio)) {
	    nSelect++;
	    size_t ibin = _series.getBin(Time(gps_s, gps_ns));
	    if (_no_retrigger) dvi[ibin] = 1;
	    else                    dvi[ibin]++;
	    _total++;
	 }
      }
   }
   cout << "triggers read:       " << nRead << endl;
   cout << "triggers in time:    " << nTime << endl;
   cout << "triggers selected:   " << nSelect << endl;
}

//======================================  List trigger selection criteria
std::string
triggers::select_string(void) const {
   size_t N = _trig_select_list.size();
   if (!N) return string("none");
   //-----------------------------------  Build selection string
   ostringstream crit;
   for (size_t i=0; i<N; i++) {
      const trig_select& tsi = _trig_select_list[i];
      if (tsi.type  == trig_select::s_pass) continue;
      if (i) crit << " && ";
      switch (tsi.type) {
      case trig_select::s_string:
	 crit << tsi.field << " == " << tsi.string;
	 break;
      case trig_select::s_num_ge:
	 crit << tsi.field << " >= " << tsi.number;
	 break;
      case trig_select::s_num_lt:
	 crit << tsi.field << " < " << tsi.number;
	 break;
      default:
	 break;
      }
   }
   return crit.str();
}


//=======================================  Test trigger selection criteria
bool
triggers::select_trigger(const xsil::MetaIO& mio) const {
   bool sel = true;
   size_t N = _trig_select_list.size();
   for (size_t i=0; i<N; i++) {
      const trig_select& tsi = _trig_select_list[i];
      const char* fname = tsi.field.c_str();
      switch (tsi.type) {
      case trig_select::s_string:
	 if (mio.getString(fname) != tsi.string) sel = false;
	 break;
      case trig_select::s_num_ge:
	 if (mio.getFloat(fname) < tsi.number) sel = false;
	 break;
      case trig_select::s_num_lt:
	 if (mio.getFloat(tsi.field.c_str()) >= tsi.number) sel = false;
	 break;
      default:
	 break;
      }
   }
   return sel;
}

//=======================================  Set debug level
void
triggers::set_debug(int level) {
   _debug = level;
}

//=======================================  Set debug level
void
triggers::set_no_retrigger(bool no_retrig) {
   _no_retrigger = no_retrig;
}

//=======================================  Set trigger table name
void 
triggers::set_table_name(const std::string& tab) {
   _table_name = tab;
}

//=======================================  Set trigger gps second field name
void 
triggers::set_time_field(const std::string& name) {
   _time_field = name;
}

//=======================================  Set trigger nanosecond fiels name
void 
triggers::set_time_ns_field(const std::string& name) {
   _time_ns_field = name;
}
  
//=======================================  Veto triggers
void
triggers::veto_stats(void) {
   cout.unsetf(ios::fixed);
   cout << "triggers vetoed:     " << _nVetoed << endl;
   _total -= _nVetoed;
   cout << "total triggers:      " << _total << endl;
}

//=======================================  Veto triggers
void
triggers::veto_triggers(const std::string& name, const TSeries& data) {
   size_t Nv = _trig_veto_list.size();
   cout << setprecision(3) << fixed;
   for (size_t iv=0; iv<Nv; iv++) {
      if (name != _trig_veto_list[iv].channel) continue;

      double veto_min = _trig_veto_list[iv].val_min;
      double veto_max = _trig_veto_list[iv].val_max;
      cout << "veto triggers with " << veto_min << " <= "
	   << _trig_veto_list[iv].channel  << " < " << veto_max << endl;
      size_t jCol = 0;
      size_t Ne = data.getNSample();
      DVectI& trigDv(*dynamic_cast<DVectI*>(_series.refDVect()));
      for (size_t i=0; i<Ne; i++) {
	 if (!trigDv[i]) continue;
	 double x = data.getDouble(i);
	 if (x >= veto_min && x < veto_max) {
	    if (!jCol) cout << "vetoed:";
	    cout << " " << data.getBinT(i).totalS();
	    if (++jCol == 5) {
	       cout << endl;
	       jCol = 0;
	    }
	    _nVetoed += trigDv[i];
	    trigDv[i] = 0;
	 }
      }
      if (jCol) cout << endl;
   }
}

//======================================  Print trigger selection rule
void 
triggers::trig_select::critter(std::ostream& out) const {
   out << "field: " << field;
   switch (type) {
   case s_pass:
      out << " pass";
      break;
   case s_string:
      out << " equal to \"" << string;
      break;
   case s_num_ge:
      out << " >= " << number;
      break;
   case s_num_lt:
      out << " <  " << number;
      break;
   default:
      out << " undefined operation.";
   }
   out << endl;
}
